/* Copyright (C) 2002 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: dirclient.c,v $
   Revision 1.2  2002/08/12 17:10:47  wb6ymh
   Made TimeNow a global.
   Removed argument to GetTimeNow().

   Revision 1.1.1.1  2002/08/10 20:33:41  wb6ymh
   initial import

*/

#include "common.h"

#ifndef _WIN32
   // FreeBSD, Linux, etc..
   #include <stdio.h>
   #include <stdlib.h>
   #include <sys/types.h>
   #ifdef TIME_WITH_SYS_TIME
      #include <sys/time.h>
      #include <time.h>
   #else
      #ifdef HAVE_SYS_TIME_H
         #include <sys/time.h>
      #else
         #include <time.h>
      #endif
   #endif
   #include <unistd.h>
   #include <sys/socket.h>
   #include <netinet/in.h>
   #include <arpa/inet.h>
   #include <errno.h>
   #include <string.h>
#ifdef HAVE_FCNTL_H
   #include <fcntl.h>
#endif
   #include <netdb.h>
#else
   // Windoze
   #include <stdio.h>
   #include <time.h>
   #include <io.h>
   #include <winsock2.h>
#endif

#include "avl.h"
#include "main.h"
#include "dirclient.h"
#include "configvars.h"
#include "ilink.h"
#include "users.h"

// This is the client version number reported to the directory server
// Does it matter ?  Beats me ...
#define PULLER_VERSION  "3.35"

int PullerStatusBusy = 0;

typedef struct {
   SERVER_REQUEST Req;
   char  Callsign[MAX_CALL_LEN+1];
   char  Qth[MAX_QTH_LEN+1];
   int   NodeID;
   int   LocalHr;
   int   LocalMin;
   int   Stations;
   int   Flag;
      #define PULLER_FLAG_IGNORE   1
      #define PULLER_FLAG_ABORT    2
      #define PULLER_FLAG_BUSY     4
} PullerInfo;

PullerInfo *pPullerDebug = NULL;

int PullerLogin(ClientInfo *pClient);
int PullerLoginAck(ClientInfo *pClient);
int PullerConnected(ClientInfo *pClient);
void ServerRequest(SERVER_REQUEST Req);
int GetStationList(ClientInfo *pClient);
int ParseStationList(ClientInfo *pClient);
void CleanUserList(void);
int PullerHandler(ClientInfo *pPuller);
void PullerCleanup(ClientInfo *pPuller);

void ServerRequest(SERVER_REQUEST Req)
{
   int i;
   ClientInfo *pClient;
   PullerInfo *pPuller;
   struct hostent *pTemp;

   pClient = CreateNewClient();
   if(pClient != NULL) {
      pPuller = malloc(sizeof(PullerInfo));
      if(pPuller != NULL) {
         pClient->p = pPuller;
         pClient->Cleanup = PullerCleanup;
         memset(pPuller,0,sizeof(PullerInfo));
         pPuller->Req = Req;

         pClient->Socket = socket(AF_INET,SOCK_STREAM,0);
         if(pClient->Socket == SOCKET_ERROR) {
            LOG_ERROR(("ServerRequest(): Socket() failed %m"));
            DeleteClient(pClient);
            pClient = NULL;
         }
      }
   }
      
   if(pClient != NULL) {
   // Set the socket to non-blocking
   #ifdef _WIN32
      i = 1;
      ioctlsocket(pClient->Socket,FIONBIO,&i);
   #else
      fcntl(pClient->Socket,F_SETFL,O_NONBLOCK);
   #endif
      
      pClient->HisAdr.i.sin_family = AF_INET;
      pClient->HisAdr.i.sin_port = htons(ILINK_DIRSERVER_PORT);
   
      pTemp = gethostbyname(DirServerHost);
      if(pTemp != NULL) {
         pClient->HisAdr.i.sin_addr.s_addr = *((u_long *) pTemp->h_addr_list[0]);
      }
      else {
      // That didn't work.  On some stacks a numeric IP address
      // will not parse with gethostbyname.  Try to convert it as a
      // numeric address
         pClient->HisAdr.i.sin_addr.s_addr = inet_addr(DirServerHost);
      }
   
      avl_insert(ClientTree,pClient);
      if(pClient->HisAdr.i.sin_addr.s_addr == INADDR_NONE) {
         LOG_ERROR(("ServerRequest(): Couldn't convert \"%s\" to an IP address.\n",
                  DirServerHost));
         DeleteClient(pClient);
         pClient = NULL;
      }
   }

   if(pClient != NULL) {
      pClient->State = PullerConnected;
      pClient->NextState = PullerLogin;

      i = connect(pClient->Socket,&pClient->HisAdr.s,sizeof(pClient->HisAdr));

      if(i == SOCKET_ERROR) {
         if(ERROR_CODE == ERROR_IN_PROGRESS) {
         // The expected result, connect takes a while
            D2PRINTF(("ServerRequest(), connect() returned ERROR_IN_PROGRESS.\n"));
            SET_WAIT4_WR(pClient);
         }
         else {
         // Some kind of error other than in progress
            DPRINTF(("ServerRequest(), connect() returned error %d.\n",ERROR_CODE));
            OpComplete(pClient,ERROR_CODE);
         }
      }
      else {
      // That certainly was quick !
         DPRINTF(("ServerRequest(), connect() didn't return an error(!).\n"));
         PullerConnected(pClient);
      }
   }
}

int PullerConnected(ClientInfo *pClient)
{
   int i;
   int Err;
   PullerInfo *pPuller = (PullerInfo *) pClient->p;
   int Ret = FALSE;

   i = connect(pClient->Socket,&pClient->HisAdr.s,sizeof(pClient->HisAdr));

#ifdef _WIN32

// Apparently there's a bug Microsoft's Windows 2000 TCP/IP stack (shock !) 
// that causes the first call to connect() after select() has indicated that
// the socket is ready for write to return an error other than WSAEISCONN.  
// The second call to connect() always *appears* to work correctly.  
// Heavy sigh...  
//
// If we're running under WinDoze try the connect a second time if (when) it
// fails.

   if(i == SOCKET_ERROR && (Err = ERROR_CODE) != ERROR_CONNECTED) {
      D2PRINTF(("PullerConnected(): connect() returned %d for client %u "
               "(first try).\n",Err,pClient->SN));
      i = connect(pClient->Socket,&pClient->HisAdr.s,sizeof(pClient->HisAdr));
   }
#endif

// BSD returns an error with an error code of EISCONN here.  That agrees
// with the man page description of connect().
//
// Linux does not return an error.  The Linux man page (at least Red Hat's) 
// doesn't include any details on the expected behavour of connect when used
// with non-blocking sockets and select as we are doing here.
//
   if(i != 0 && (Err = ERROR_CODE) != ERROR_CONNECTED) {
   // Connection failed.
      DPRINTF(("PullerConnected(): connect() returned %d for client %u.\n",
               Err,pClient->SN));
      DeleteClient(pClient);
   }
   else {
      switch(pPuller->Req) {
         case SERV_REQ_LOGIN:
         case SERV_REQ_LOGOUT:
         case SERV_REQ_SHUTDOWN:
         case SERV_REQ_LOGIN_AND_LIST:
         // Login into the server
            Ret = PullerLogin(pClient);
            break;

         case SERV_REQ_STATION_LIST:
         // Refresh server list
            Ret = GetStationList(pClient);
            break;

         default:
            LOG_ERROR(("PullerConnected(), unexpected request %d!\n",
                     pPuller->Req));
            DeleteClient(pClient);
            break;
      }
   }

   return Ret;
}

int PullerLogin(ClientInfo *pClient)
{
   int WriteSize;
   char c = 'l';
   struct tm *tm = localtime(&TimeNow.tv_sec);
   int Ret = 0; 
   PullerInfo *pPuller = (PullerInfo *) pClient->p;
   char *Status;
   
   pClient->BufSize = 256;
   pClient->Buf = (char *) malloc(pClient->BufSize);

   if((WRITE(pClient->Socket,&c,1) != 1)) {
   // This isn't good.  It's the first thing we've sent.
   // The buffer can't be full yet !
      DPRINTF(("PullerLogin(): write() failed for client %u, error = %d.\n",
               pClient->SN,ERROR_CODE));
      OpComplete(pClient,ERROR_CODE);
   }
   else {
      switch(pPuller->Req) {
         default:
            LOG_ERROR(("PullerLogin(): Unknown server request"));

         case SERV_REQ_LOGIN:
         case SERV_REQ_LOGIN_AND_LIST:
            Status = PullerStatusBusy ? "BUSY" : "ONLINE";
            break;

         case SERV_REQ_SHUTDOWN:
         case SERV_REQ_LOGOUT:
            Status = "OFF-";
            break;
      }
      WriteSize = snprintf(pClient->Buf,pClient->BufSize,
                           "%s%c%c%s\r%s%s(%d:%2d)\r%s\r",
                           ConferenceCall,0xac,0xac,ConferencePass,Status,
                           PULLER_VERSION,tm->tm_hour,tm->tm_mday,ConferenceQth);
      if(WriteSize == pClient->BufSize) {
         LOG_ERROR(("PullerLogin(): Temp buffer too small !\n"));
         DeleteClient(pClient);
      }
      else {
         pClient->BufSize = WriteSize;
         pClient->Count = 0;
         pClient->State = SendBuffer;
         pClient->NextState = PullerLoginAck;
         Ret = 1;
      }
   }

   return Ret;
}

int PullerLoginAck(ClientInfo *pClient)
{
   int  BytesRead;
   char Data;
   PullerInfo *pPuller = (PullerInfo *) pClient->p;

   if(pClient->BufSize > 2) {
   // We've sent our info, now wait for the response from the server
      pClient->BufSize = 2;
      pClient->Count = 0;
   }

   while(pClient->Count < 2) {
      if((BytesRead = READ(pClient->Socket,&Data,1)) == 1) {
            pClient->Buf[pClient->Count++] = Data;
      }
      else if(BytesRead == 0) {
      // Connection has closed
         OpComplete(pClient,ERROR_CODE);
         break;
      }
      else if(ERROR_CODE == ERROR_WOULD_BLOCK) {
      // Nothing available to read now, keep waiting
         SET_WAIT4_RD(pClient);
         break;
      }
      else {
      // Some kind of error other than no data available
         DPRINTF(("PullerLoginAck(), READ() returned error %d.\n",ERROR_CODE));
         OpComplete(pClient,ERROR_CODE);
      }
   }

   if(pClient->Count == 2) {
   // If we succeeded in loging in our buffer should contain "ok"
      pClient->Buf[2] = 0;
      if(strcmp(pClient->Buf,"OK") == 0) {
         DPRINTF(("PullerLoginAck(): Client %u successfully updated status.\n",
                  pClient->SN));
         switch(pPuller->Req) {
            case SERV_REQ_LOGIN:
            // We're done
               OpComplete(pClient,0);
               break;

            case SERV_REQ_LOGIN_AND_LIST:
            // Now list the stations
               DeleteClient(pClient);
               ServerRequest(SERV_REQ_STATION_LIST);
               break;

            case SERV_REQ_SHUTDOWN:
               LOG_NORM(("Logged out, exiting."));
               bRunning = FALSE;
               break;

            default:
               LOG_ERROR(("PullerLoginAck(), unexpected request %d!\n",
                        pPuller->Req));
               DeleteClient(pClient);
               break;
         }
      }
      else {
         LOG_ERROR(("PullerLoginAck(): login error server returned \"%s\" for"
                  "client %u.\n",pClient->Buf,pClient->SN));
      }
   }

   return 0;
}

int GetStationList(ClientInfo *pClient)
{
   char c = 's';
   PullerInfo *pPuller = (PullerInfo *) pClient->p;
   
   pPuller->Req = SERV_REQ_STATION_LIST;
   pClient->BufSize = 256;
   pClient->Buf = (char *) malloc(pClient->BufSize);
   pClient->Count = 0;
   pClient->Terminator = '\n';
   pClient->State = ReadLine;
   pClient->NextState = ParseStationList;
   SET_WAIT4_RD(pClient);

   if((WRITE(pClient->Socket,&c,1) != 1)) {
   // This isn't good.  It's the first thing we've sent.
   // The buffer can't be full yet !
      DPRINTF(("GetStationList(): write() failed, error = %d.\n",ERROR_CODE));
      DeleteClient(pClient);
   }

   return 0;
}

int ParseStationList(ClientInfo *pClient)
{
   PullerInfo *pPuller = (PullerInfo *) pClient->p;
   UserInfo *pUser;
   UserInfo UserLookup;
   char *cp;
   int Ret = 0;
   int i;

   if(pClient->Err != 0) {
   // That's it folks, were done.
      pPuller->Flag = PULLER_FLAG_ABORT;
      if(pClient->Count >= 3) {
         if(strncmp(pClient->Buf,EndOfData,strlen(EndOfData)) == 0) {
            CleanUserList();
            DPRINTF(("ParseStationList(): completed successfully, "
                     "%d stations listed.\n",pPuller->Stations));
         }
         else {
         // Unexpected disconnection/socket error
            DPRINTF(("ParseStationList(): Unexpected disconnection.\n"));
         }
      }
   }
   else if(!(pPuller->Flag & PULLER_FLAG_IGNORE)) {
      switch(pClient->x) {
         case 0:  // Start of data
            if(strncmp(pClient->Buf,StartOfData,strlen(StartOfData)) != 0) {
               LOG_ERROR(("ParseStationList(): Expected StartOfData got \"%s\".\n",
                        pClient->Buf));
               pPuller->Flag = PULLER_FLAG_ABORT;
            }
            break;

         case 1:  // Number of stations, just ignore it
            break;

         case 2:  // Callsign
            if(pClient->Count > MAX_CALL_LEN) {
               LOG_ERROR(("ParseStationList(): Callsign \"%s\" is too long.\n",
                        pClient->Buf));
               pPuller->Flag = PULLER_FLAG_IGNORE;
            }
            else if(pClient->Buf[0] == ' ') {
               // Message or something ignore it
               pPuller->Flag = PULLER_FLAG_IGNORE;
            }
            else {
               strcpy(pPuller->Callsign,pClient->Buf);
            }
            break;

         case 3:  // Qth / status
            if((cp = strchr(pClient->Buf,'[')) != NULL) {
               *cp++ = 0;
               i = sscanf(cp,"%*s %d:%d",&pPuller->LocalHr,&pPuller->LocalMin);
               if(i != 2) {
                  pPuller->LocalHr = -1;
               }

               if(strncmp(cp,"BUSY",4) == 0) {
                  pPuller->Flag = PULLER_FLAG_BUSY;
               }
               else if(strncmp(cp,"ON",2) != 0) {
               // Invalid entry, unknown state
                  pPuller->Flag = PULLER_FLAG_IGNORE;
                  break;
               }
            }
            else {
            // Invalid entry, no status
               pPuller->Flag = PULLER_FLAG_IGNORE;
               break;
            }

            if(strlen(pClient->Buf) > MAX_QTH_LEN) {
               LOG_ERROR(("ParseStationList(): Qth \"%s\" is too long.\n",
                        pClient->Buf));
               pPuller->Flag = PULLER_FLAG_IGNORE;
            }
            else {
               strcpy(pPuller->Qth,pClient->Buf);
            }
            break;

         case 4:  // NodeID
            if(sscanf(pClient->Buf,"%d",&pPuller->NodeID) != 1) {
               LOG_ERROR(("ParseStationList(): can't convert NodeID \"%s\".\n",
                        pClient->Buf));
               pPuller->Flag = PULLER_FLAG_IGNORE;   // Ignore this guy
            }
            break;

         case 5:  // IP address
            UserLookup.Callsign = pPuller->Callsign;
            if((pUser = avl_find(UserTree,&UserLookup)) == NULL) {
            // New user
               D2PRINTF(("ParseStationList(): Adding new user \"%s\".\n",
                        pPuller->Callsign));
               pUser = CreateNewUser(pPuller->Callsign);
               avl_insert(UserTree,pUser);
            }
            else {
               D2PRINTF(("ParseStationList(): Updating user \"%s\".\n",
                        pPuller->Callsign));
            }

            if(pUser != NULL) {
               pUser->bActive = pUser->bRemote = pUser->bRefreshed = TRUE;
               if(pPuller->Flag & PULLER_FLAG_BUSY) {
                  pUser->bBusy = TRUE;
               }
               else {
                  pUser->bBusy = FALSE;
               }

               if(pUser->Qth != NULL) {
                  free(pUser->Qth);
               }
               pUser->Qth = strdup(pPuller->Qth);

               if(pUser->HisAdr != NULL) {
                  free(pUser->HisAdr);
               }
               pUser->HisAdr = strdup(pClient->Buf);

               if(pUser->NodeID != NULL) {
                  free(pUser->NodeID);
               }
               sprintf(pPuller->Qth,"%d",pPuller->NodeID);
               pUser->NodeID = strdup(pPuller->Qth);

               pUser->LocalHr = pPuller->LocalHr;
               pUser->LocalMin = pPuller->LocalMin;
            }
            break;
      }
   }

   if(pPuller->Flag & PULLER_FLAG_ABORT) {
      DeleteClient(pClient);
   }
   else {
      if(pClient->x++ == 5) {
         if(!(pPuller->Flag & PULLER_FLAG_IGNORE)) {
            pPuller->Stations++;
         }
      // Back to reading expecting a callsign
         pClient->x = 2;
         pPuller->Flag = 0;   // parse entry again
      }
      pClient->Count = 0;
      pClient->State = ReadLine;
      pClient->NextState = ParseStationList;
      Ret = 1;
   }
   
   return Ret;
}

// Remove any stations from the user list that are no longer present
// in the list we just pulled from the pull server
void CleanUserList()
{
   UserInfo *pUser;
   UserInfo *pNextUser;
   struct avl_traverser avl_trans;

   pNextUser = (UserInfo *) avl_t_first(&avl_trans,UserTree);
   while((pUser = pNextUser) != NULL) {
   // NB: Get the next user now in case we deleted the current user
      pNextUser = (UserInfo *) avl_t_next(&avl_trans);

      if(pUser->bRemote) {
         if(!pUser->bRefreshed){
         // He's gone
            DeleteUser(pUser,FALSE);
         }
         else {
            pUser->bRefreshed = FALSE;
         }
      }
   }
}

void PullerCleanup(ClientInfo *p)
{
   free(p->p);
   p->p = NULL;
}


